<h3>Downloads</h3>
<p>Current version: <b>{{ release }}</b></p>
<p>Downloads:
<a href="http://pypi.python.org/pypi/asteval/">PyPI (Python.org)</a>

<p> try <tt>pip install asteval</tt>

<p>
Development version: <br> &nbsp; &nbsp; <a href="https://github.com/newville/asteval/">github.com</a>

<hr>
